------------------------------------------------------------------------------
--                                                                          --
--                           GNAT RAVENSCAR for NXT                         --
--                                                                          --
--                       Copyright (C) 2010, AdaCore                        --
--                                                                          --
-- This is free software; you can  redistribute it  and/or modify it under  --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. This is distributed in the hope that it will be useful, but WITH-  --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNARL; see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable  to  be --
-- covered  by the  GNU  General  Public  License.  This exception does not --
-- however invalidate  any other reasons why  the executable file  might be --
-- covered by the  GNU Public License.                                      --
--                                                                          --
------------------------------------------------------------------------------

with Interfaces; use Interfaces;
with Nxt.Twi; use Nxt.Twi;
with Ada.Unchecked_Conversion;

package body Nxt.Avr is

   type Avr_Msg is array (Natural range <>) of Unsigned_8;
   subtype Avr_Msg_Out is Avr_Msg (0 .. 8);
   subtype Avr_Msg_In is Avr_Msg (0 .. 12);
   --  Messages type.

   Avr_Address : constant Device_Id := 1;

   Out_Msg : Avr_Msg_Out := (others => 0);
   In_Msg : Avr_Msg_In;
   --  Current messages.

   Avr_String : constant String :=
     Character'Val (16#Cc#) & "Let's samba nxt arm in arm, (c)LEGO System A/S";
   --  Hello string.

   procedure Init is
   begin
      Nxt.Twi.Send (Avr_Address, 0, Avr_String'Address, Avr_String'Length);
   end Init;

   procedure Send_Frame is
      Out_Msg2 : Avr_Msg_Out;
      Sum : Unsigned_8;
   begin
      --  Make a copy and compute the checksum.
      Sum := 0;
      for I in Out_Msg'First .. Out_Msg'Last - 1 loop
         Out_Msg2 (I) := Out_Msg (I);
         Sum := Sum + Out_Msg (I);
      end loop;
      Out_Msg2 (Out_Msg2'Last) := not Sum;

      Nxt.Twi.Send (Avr_Address, 0, Out_Msg2'Address, Out_Msg2'Size / 8);
   end Send_Frame;

   procedure Power_Down is
   begin
      Out_Msg := (others => 0);
      Out_Msg (0) := 16#5a#;
      Out_Msg (1) := 0;
      --  Send_Frame;
   end Power_Down;

   procedure Set_Motor (Motor : Motor_Id; Pwm : Pwm_Type; Break : Boolean)
   is
      function To_Unsigned_8 is new Ada.Unchecked_Conversion
        (Integer_8, Unsigned_8);
   begin
      Out_Msg (1) := 8;
      Out_Msg (2 + Motor_Id'Pos (Motor)) := To_Unsigned_8 (Integer_8 (Pwm));
      if Break then
         Out_Msg (6) :=
           Out_Msg (6) and not Shift_Left (1, Motor_Id'Pos (Motor));
      else
         Out_Msg (6) :=
           Out_Msg (6) or Shift_Left (1, Motor_Id'Pos (Motor));
      end if;
   end Set_Motor;

   procedure Read_Data is
      Sum : Unsigned_8;
      In_Msg2 : Avr_Msg_In;
   begin
      Nxt.Twi.Recv (Avr_Address, 0, In_Msg2'Address, In_Msg2'Size / 8);

      --  Compute the checksum.
      Sum := 0;
      for I in In_Msg2'Range loop
         Sum := Sum + In_Msg2 (I);
      end loop;
      if Sum /= 16#Ff# then
         return;
      end if;

      --  Copy the valid message.
      In_Msg := In_Msg2;
   end Read_Data;

   function Get_Button return Button_Kind is
      Val : Unsigned_16;
   begin
      --  Decode.
      Val := Unsigned_16 (In_Msg (8))
        + Shift_Left (Unsigned_16 (In_Msg (9)), 8);

      if Val >= 16#7ff# then
         return Button_Power;
      end if;

      --   Assume only one button is pushed.
      if Val > 16#300# then
         return Button_Middle;
      elsif Val > 16#100# then
         return Button_Right;
      elsif Val > 16#60# then
         return Button_Left;
      end if;

      return Button_None;
   end Get_Button;

   function Get_Sensor (Sensor : Sensor_Id) return Unsigned_16 is
      P : Natural := 2 * Sensor_Id'Pos (Sensor);
   begin
      return Unsigned_16 (In_Msg (P))
        + Shift_Left (Unsigned_16 (In_Msg (P + 1)), 8);
   end Get_Sensor;
end Nxt.Avr;
