------------------------------------------------------------------------------
--                                                                          --
--                           GNAT RAVENSCAR for NXT                         --
--                                                                          --
--                       Copyright (C) 2010, AdaCore                        --
--                                                                          --
-- This is free software; you can  redistribute it  and/or modify it under  --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. This is distributed in the hope that it will be useful, but WITH-  --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNARL; see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable  to  be --
-- covered  by the  GNU  General  Public  License.  This exception does not --
-- however invalidate  any other reasons why  the executable file  might be --
-- covered by the  GNU Public License.                                      --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Real_Time; use Ada.Real_Time;
with Nxt.Bt; use Nxt.Bt;
with Nxt.Bc4; use Nxt.Bc4;

package body Nxt.Bluetooth is
   procedure Wait_Ms is
   begin
      delay until Clock + Milliseconds (10);
   end Wait_Ms;

   procedure Init is
   begin
      Nxt.Bt.Init;
      Nxt.Bt.Reset;
   end Init;

   subtype Name_Packet_Subtype is Packet_Type (0 .. 15);

   procedure String_To_Name (Str : String; Name : out Name_Type) is
   begin
      if Str'Length > Name'Length then
         for I in 0 .. Name'Length - 1 loop
            Name (Name'First + I) := Character'Pos (Str (Str'First + I));
         end loop;
      else
         for I in 0 .. Str'Length - 1 loop
            Name (Name'First + I) := Character'Pos (Str (Str'First + I));
         end loop;
         for I in 0 .. Name'Length - Str'Length - 1 loop
            Name (Name'First + Str'Length + I) := 0;
         end loop;
      end if;
   end String_To_Name;

   procedure Set_Friendly_Name (Name : Name_Type) is
      Cmd : Packet_Type (0 .. 16);
      Reply : Packet_Type (0 .. 16);
   begin
      Cmd (0)  := MSG_SET_FRIENDLY_NAME;
      Cmd (1 .. 16) := Name_Packet_Subtype (Name);
      Send_Packet (Cmd);

      loop
         Recv_Packet (Reply);
         if Reply (0) /= 0
           and then Reply (1) = MSG_SET_FRIENDLY_NAME_ACK
         then
            return;
         else
            Wait_Ms;
         end if;
      end loop;
   end Set_Friendly_Name;

   subtype Device_Packet_Subtype is Packet_Type (0 .. 26);
   procedure Packet_To_Dev (Packet : Device_Packet_Subtype;
                            Dev : out Device_Type) is
   begin
      Dev := (Addr => Address_Type (Packet (0 .. 6)),
              Name => Name_Type (Packet (7 .. 22)),
              Class => Class_Type (Packet (23 .. 26)));
   end Packet_To_Dev;

   procedure Dev_To_Packet (Dev : Device_Type;
                            Packet : out Device_Packet_Subtype) is
   begin
      Packet (0 .. 6) := Packet_Type (Dev.Addr);
      Packet (7 .. 22) := Packet_Type (Dev.Name);
      Packet (23 .. 26) := Packet_Type (Dev.Class);
   end Dev_To_Packet;

   procedure Get_Known_Peers (Devs : out Device_Array; Len : out Natural) is
      Cmd : Packet_Type (0 .. 0);
      Reply : Packet_Type (0 .. 31);
   begin
      Len := 0;
      Cmd (0) := MSG_DUMP_LIST;
      Send_Packet (Cmd);

      loop
         Recv_Packet (Reply);
         if Reply (0) /= 0 then
            if Reply (1) = MSG_LIST_ITEM then
               if Len < Devs'Length then
                  Packet_To_Dev (Reply (2 .. 28), Devs (Devs'First + Len));
                  Len := Len + 1;
               end if;
            elsif Reply (1) = MSG_LIST_DUMP_STOPPED then
               return;
            end if;
         else
            Wait_Ms;
         end if;
      end loop;
   end Get_Known_Peers;

   procedure Inquiry (Timeout_Sec : Natural;
                      Devs : out Device_Array; Len : out Natural) is
      Cmd : Packet_Type (0 .. 7);
      Reply : Packet_Type (0 .. 32);
   begin
      Len := 0;

      Cmd (0) := MSG_BEGIN_INQUIRY;
      if Devs'Length > 255 then
         Cmd (1) := 255;
      else
         Cmd (1) := Devs'Length;        --  Max dev
      end if;
      Cmd (2) := Unsigned_8 (Timeout_Sec / 256);        --  Timeout HI
      Cmd (3) := Unsigned_8 (Timeout_Sec mod 256);      --  Timeout LO
      Cmd (4) := 0;
      Cmd (5) := 0;
      Cmd (6) := 0;
      Cmd (7) := 0;
      Send_Packet (Cmd);

      loop
         Recv_Packet (Reply);
         if Reply (0) /= 0 then
            if Reply (1) = MSG_INQUIRY_RUNNING then
               null;
            elsif Reply (1) = MSG_INQUIRY_STOPPED then
               return;
            elsif Reply (1) = MSG_INQUIRY_RESULT then
               if Len < Devs'Length then
                  Packet_To_Dev (Reply (2 .. 28), Devs (Devs'First + Len));
                  Len := Len + 1;
               end if;
            end if;
         else
            Wait_Ms;
         end if;
      end loop;
   end Inquiry;

   procedure Set_Discoverable (On : Boolean) is
      Cmd : Packet_Type (0 .. 1);
      Reply : Packet_Type (0 .. 3);
   begin
      Cmd (0) := MSG_SET_DISCOVERABLE;
      Cmd (1) := Boolean'Pos (On);
      Send_Packet (Cmd);

      loop
         Recv_Packet (Reply);
         if Reply (0) /= 0 then
            if Reply (1) = MSG_DISCOVERABLE_ACK then
               return;
            end if;
         else
            Wait_Ms;
         end if;
      end loop;
   end Set_Discoverable;

   procedure Add_Known_Device (Dev : Device_Type; Success : out Boolean) is
      Cmd : Packet_Type (0 .. 27);
      Reply : Packet_Type (0 .. 3);
   begin
      Cmd (0) := MSG_ADD_DEVICE;
      Dev_To_Packet (Dev, Cmd (1 .. 27));
      Send_Packet (Cmd);

      loop
         Recv_Packet (Reply);
         if Reply (0) = 0 then
            Wait_Ms;
         elsif Reply (1) = MSG_LIST_RESULT then
            Success := Reply (2) = 16#50#;
            return;
         end if;
      end loop;
   end Add_Known_Device;

   Handle : Unsigned_8;

   procedure Start_Data is
      Cmd : Packet_Type (0 .. 1);
   begin
      Cmd (0) := MSG_OPEN_STREAM;
      Cmd (1) := Handle;
      Send_Packet (Cmd (0 .. 1));
      delay until Clock + Milliseconds (50);
      Mode_Data;
   end Start_Data;

   procedure Accept_Connection (Pin : Pin_Type;
                                Addr : out Address_Type; Success : out Boolean)
   is
      Cmd : Packet_Type (0 .. 0);
      Pin_Cmd : Packet_Type (0 .. 23);
      Accpt_Cmd : Packet_Type (0 .. 1);
      Reply : Packet_Type (0 .. 32);
   begin
      Cmd (0) := MSG_OPEN_PORT;
      Send_Packet (Cmd);

      loop
         Recv_Packet (Reply);
         if Reply (0) = 0 then
            Wait_Ms;
         else
            if Reply (1) = MSG_PORT_OPEN_RESULT then
               null;
            elsif Reply (1) = MSG_REQUEST_PIN_CODE then
               Pin_Cmd (0) := MSG_PIN_CODE;
               Pin_Cmd (1 .. 7) := Reply (2 .. 8);
               Pin_Cmd (8 .. 23) := Packet_Type (Pin);
               Send_Packet (Pin_Cmd);
            elsif Reply (1) = MSG_PIN_CODE_ACK then
               null;
            elsif Reply (1) = MSG_REQUEST_CONNECTION then
               Addr := Address_Type (Reply (2 .. 8));
               Accpt_Cmd (0) := MSG_ACCEPT_CONNECTION;
               Accpt_Cmd (1) := 1;
               Send_Packet (Accpt_Cmd);
            elsif Reply (1) = MSG_CONNECT_RESULT then
               if Reply (2) = 1 then
                  Handle := Reply (3);
                  Success := True;
                  Start_Data;
               else
                  Success := False;
               end if;
               return;
            end if;
         end if;
      end loop;
   end Accept_Connection;

   procedure Connect (Addr : Address_Type; Success : out Boolean) is
      Cmd : Packet_Type (0 .. 7);
      Reply : Packet_Type (0 .. 3);
   begin
      Cmd (0) := MSG_CONNECT;
      Cmd (1 .. 7) := Packet_Type (Addr);
      Send_Packet (Cmd);

      loop
         Recv_Packet (Reply);
         if Reply (0) = 0 then
            Wait_Ms;
         elsif Reply (1) = MSG_CONNECT_RESULT then
            if Reply (2) = 1 then
               Handle := Reply (3);
               Success := True;
               Start_Data;
            else
               Success := False;
            end if;
            return;
         elsif Reply (1) = MSG_CLOSE_CONNECTION_RESULT then
            Success := False;
            return;
         end if;
      end loop;
   end Connect;
end Nxt.Bluetooth;
